MODULE Equations;

 (*
    Equations - Equation evaluation.       
    Copyright (C) 1996 Michael Griebling
 
    This module is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
 
    This module is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*)

IMPORT XI:= ExIntegers, X:= ExNumbers, el:= UnixExecLists,
       XMO:= ExMathLib0, XM:= ExComplexLib0, f:= Files,
       s:= Strings, p:= UnixPlot, V:= Variables, set:= Settings,
       r:= UnixRegister, uc:= UnitConversions, Rts:=Termination, 
       sc:= Scanner, Out, cli:=CLIErrors, Statistics,
       tr:=TextRider;

VAR
  Token: INTEGER;
  ResultStr, CommandLine : sc.String;
  Equations*, Results* : el.List;
  EquationFile, ResultFile: ARRAY 18 OF CHAR;


(* $CopyArrays- *)
PROCEDURE StoreVariable(Location: ARRAY OF CHAR; Value : XM.ExComplexType);
(* Store the `Value' argument in the `Location' variable. *)
VAR ok: BOOLEAN;
BEGIN 
  V.Set(Location, Value, ok);
  IF ~ok THEN sc.Mark(cli.TooManyVariables) END
END StoreVariable;


(* $CopyArrays- *)
PROCEDURE RecallVariable(Location : ARRAY OF CHAR; VAR Value : XM.ExComplexType);
(* Recall the contents of the `Location' variable and return *)
VAR ok: BOOLEAN;
BEGIN 
  V.Get(Location, Value, ok);
  IF ~ok THEN sc.Mark(X.UndefinedStorage) END
END RecallVariable;


PROCEDURE^ Expression (VAR Result : XM.ExComplexType);


PROCEDURE PlotFx (x : LONGREAL) : LONGREAL;
VAR xConv : XM.ExComplexType; ok : BOOLEAN;
BEGIN
  XM.LongRealToComplex(x, xConv);    (* convert argument to ExNumber *)
  V.Set("M0", xConv, ok);            (* store the argument in M0 *)
  COPY(CommandLine, ResultStr);      (* get a new equation *)
  sc.Init(ResultStr); sc.Get(Token); (* prime the tokenizer *)
  Expression(xConv);                 (* evaluate the expression *)
  RETURN XM.ComplexToLongReal(xConv) (* return the result *)
END PlotFx;


PROCEDURE Print (name, arg: ARRAY OF CHAR);
BEGIN
  Out.String(name); Out.String(" = "); Out.String(arg); Out.Ln
END Print;


PROCEDURE Help;
BEGIN
  Out.Ln;
  Out.String("    +            Addition"); Out.Ln;
  Out.String("    -            Subtraction"); Out.Ln;
  Out.String("    *,          Multiplication"); Out.Ln;
  Out.String("    /,          Division"); Out.Ln;
  Out.String("                Squared"); Out.Ln;
  Out.String("                Cubed"); Out.Ln;
  Out.String("    -,        Reciprocal"); Out.Ln;
  Out.String("    ()           Brackets"); Out.Ln;
  Out.String("    ^, **        Power"); Out.Ln;
  Out.String("    %            x 0.01"); Out.Ln;
  Out.String("    &, And       Logical And"); Out.Ln;
  Out.String("    |, Or        Logical Inclusive Or"); Out.Ln;
  Out.String("    Xor          Logical Exclusive Or"); Out.Ln;
  Out.String("    ~,Not        Logical Complement"); Out.Ln;
  Out.String("    Mod          Modulo"); Out.Ln;
  Out.String("    Div          Integer Division"); Out.Ln;
  Out.String("    Sqrt         Square Root"); Out.Ln;
  Out.String("    Cbrt         Cube Root"); Out.Ln;
  Out.String("    Root         Any Root"); Out.Ln;
  Out.String("    Abs          Absolute value"); Out.Ln;
  Out.String("    Rand         Random number between 0 and 1"); Out.Ln;
  Out.String("    e            Natural Log Base"); Out.Ln;
  Out.String("    e^           Power of e"); Out.Ln;
  Out.String("    i,          Imaginary number"); Out.Ln;
  Out.String("                Complex angle"); Out.Ln;
  Out.String("    Im           Imaginary part of a number"); Out.Ln;
  Out.String("    Re           Real part of a number"); Out.Ln;
  Out.String("               Polar to rectangle conversion"); Out.Ln;
  Out.String("    r            Complex magnitude"); Out.Ln;
  Out.String("    Conj         Complex conjugate"); Out.Ln;
  Out.String("    Conv <n>     Convert n units to destination units"); Out.Ln;
  Out.String("    nCr          Factorial combinations"); Out.Ln;
  Out.String("    nPr          Factorial permutations"); Out.Ln;
  Out.String("    Frac         Fractional part of a number"); Out.Ln;
  Out.String("    Int          Integer part of a number"); Out.Ln;
  Out.String("    Sign         Sign of a number (-1 or 1)"); Out.Ln;
  Out.String("    Plot <exp>   Plot the expression as f(M0) (not supported -- yet)"); Out.Ln;
  Out.String("    Min(<x>;<y>) Minimum of x and y"); Out.Ln;
  Out.String("    Max(<x>;<y>) Maximum of x and y"); Out.Ln;
  Out.String("    Ln           Natural Logarithm"); Out.Ln;
  Out.String("    Log          Base 10 Logarithm"); Out.Ln;
  Out.String("    Sin, ASin    Sine, Arcsine"); Out.Ln;
  Out.String("    Cos, ACos    Cosine, Arccosine"); Out.Ln;
  Out.String("    Tan, ATan    Tangent, Arctangent"); Out.Ln;
  Out.String("    Sinh, ASinh  Hyperbolic Sine, Arcsine"); Out.Ln;
  Out.String("    Cosh, ACosh  Hyperbolic Cosine, Arccosine"); Out.Ln;
  Out.String("    Tanh, ATanh  Hyperbolic Tangent, Arctangent"); Out.Ln;
  Out.String("    SBit         Set Bit"); Out.Ln;
  Out.String("    CBit         Clear Bit"); Out.Ln;
  Out.String("    TBit         Toggle Bit"); Out.Ln;
  Out.String("    Shr          Shift Right"); Out.Ln;
  Out.String("    Shl          Shift Left"); Out.Ln;
  Out.String("    Asr          Arithmetic Shift Right"); Out.Ln;
  Out.String("    Ror          Rotate Right"); Out.Ln;
  Out.String("    Rol          Rotate Left"); Out.Ln;
  Out.String("    <var>        Variable contents"); Out.Ln;
  Out.String("    ST <var>     Store to variable"); Out.Ln;
  Out.String("    DEL <var>    Deletes the variable"); Out.Ln;
  Out.String("    List         Lists all defined variables"); Out.Ln;
  Out.String("    Help, help   Displays this list"); Out.Ln;
  Out.String("    Pi           Constant Pi"); Out.Ln;
  Out.String("    SCI          Toggle Scientific/Floating Point"); Out.Ln;
  Out.String("    BAS n        Change to Base n"); Out.Ln;
  Out.String("    DIG n        Use n Digits"); Out.Ln;
  Out.String("    DP n         Use n Decimal Places"); Out.Ln;
  Out.String("    DRG          Toggle Degree/Radian/Grad"); Out.Ln;
END Help;


PROCEDURE Factor (VAR Result : XM.ExComplexType);
VAR
  SaveBase : XI.BaseType;
  tmp     : XM.ExComplexType;
  t        : X.ExNumType;

  PROCEDURE Next;
  BEGIN
    sc.Get(Token); Factor(Result);
  END Next;

  PROCEDURE Comma;
  BEGIN
    sc.Get(Token);
    IF Token#sc.LeftBrace THEN sc.Mark(X.MismatchBraces) END;
    sc.Get(Token); Factor(tmp);
    IF Token#sc.Semi THEN sc.Mark(X.IllegalOperator) END;
    sc.Get(Token); Factor(Result);
    IF Token#sc.RightBrace THEN sc.Mark(X.MismatchBraces) END;
    sc.Get(Token)
  END Comma;

  PROCEDURE Fix;
  BEGIN
    XM.Init(tmp, t, X.Ex0)
  END Fix;

  PROCEDURE FixR;
  BEGIN
    XM.Init(Result, t, X.Ex0)
  END FixR;

BEGIN
  CASE Token OF
      sc.LeftBrace  : sc.Get(Token); Expression(Result);
                      IF Token = sc.RightBrace THEN sc.Get(Token);
                      ELSE sc.Mark(X.MismatchBraces)
                      END;
    | sc.Number     : sc.Get(Token); Result := sc.val;
                      IF Token = sc.Number THEN sc.Mark(X.IllegalNumber) END
    | sc.Pi         : sc.Get(Token); XM.Init(Result, X.pi, X.Ex0)
    | sc.Complement : Next(); XI.ExOnesComp(t, Result.re); FixR
    | sc.Sin        : Next(); XM.sinX(Result, Result);
    | sc.Cos        : Next(); XM.cosX(Result, Result);
    | sc.Tan        : Next(); XM.tanX(Result, Result);
    | sc.ArcSin     : Next(); XM.arcsinX(Result, Result);
    | sc.ArcCos     : Next(); XM.arccosX(Result, Result);
    | sc.ArcTan     : Next(); XM.arctanX(Result, Result);
    | sc.Sinh       : Next(); XM.sinhX(Result, Result);
    | sc.Cosh       : Next(); XM.coshX(Result, Result);
    | sc.Tanh       : Next(); XM.tanhX(Result, Result);
    | sc.ArcSinh    : Next(); XM.arcsinhX(Result, Result);
    | sc.ArcCosh    : Next(); XM.arccoshX(Result, Result);
    | sc.ArcTanh    : Next(); XM.arctanhX(Result, Result);
    | sc.SquareRoot : Next(); XM.sqrtX(Result, Result);
    | sc.CubeRoot   : Next(); X.ExNumb(3, 0, 0, t); Fix; XM.rootX(Result, Result, tmp);
    | sc.NaturalLog : Next(); XM.lnX(Result, Result);
    | sc.Log        : Next(); XM.logX(Result, Result);
    | sc.PowerOfe   : Next(); XM.expX(Result, Result);
    | sc.Var        : sc.Get(Token); RecallVariable(sc.var, Result);
    | sc.Base       : SaveBase := set.State.LocalBase;
                      set.State.LocalBase := 10;
                      Next();
                      set.State.LocalBase := SHORT(SHORT(X.ExToLongInt(Result.re)));
                      IF (set.State.LocalBase < 2) OR
                         (set.State.LocalBase > 16) THEN
                        set.State.LocalBase := SaveBase;
                      END;
                      Result := set.State.LastAnswer;
    | sc.Digits     : Next();
                      IF X.ExStatus = X.Okay THEN
                        set.State.NumbDigits := SHORT(X.ExToLongInt(Result.re));
                        X.SetMaxDigits(set.State.NumbDigits);
                        set.State.NumbDigits := X.GetMaxDigits();
                        Result := set.State.LastAnswer;
                      END;
    | sc.Decimals   : Next();
                      IF X.ExStatus = X.Okay THEN
                        set.State.DecPoint := SHORT(X.ExToLongInt(Result.re));
                        Result := set.State.LastAnswer;
                      END;
    | sc.Notation   : sc.Get(Token);
                      set.State.Notation:=(set.State.Notation+1) MOD 3;
                      Result := set.State.LastAnswer;
    | sc.DegRadGrad : sc.Get(Token);
                      IF XMO.AngleUnit = XMO.Gradians THEN
                        XMO.AngleUnit := XMO.Degrees;
                      ELSE INC(XMO.AngleUnit) END;
                      Result := set.State.LastAnswer;
    | sc.ListVar    : sc.Get(Token);
                      V.Iterate(Print);
                      Result := XM.Ex0
    | sc.Help       : sc.Get(Token);
                      Help;
                      Result := XM.Ex0
    | sc.Plot       : s.Delete(CommandLine, 0, 4);      (* extract plot equation *)
                      p.PlotFunction(PlotFx);
                      Result := XM.Ex0; Token := sc.Empty
    | sc.Delete     : sc.Get(Token); 
                      IF Token=sc.Var THEN V.Delete(sc.var) 
                      ELSE sc.Mark(cli.IllegalVariable)
                      END;
                      Result := XM.Ex0; sc.Get(Token)                      
    | sc.iToken     : sc.Get(Token); XM.Init(Result, X.Ex0, X.Ex1)
    | sc.rToken     : Next(); XM.PolarMag(t, Result); FixR
    | sc.Theta      : Next(); XM.PolarAngle(t, Result); FixR
    | sc.ImagPart   : Next(); XM.Init(Result, Result.im, X.Ex0)
    | sc.RealPart   : Next(); XM.Init(Result, Result.re, X.Ex0)
    | sc.Convert    : Next(); uc.Convert(Result)
    | sc.IntPart    : Next(); t:=Result.re; X.ExTrunc(t); FixR
    | sc.FracPart   : Next(); t:=Result.re; X.ExFrac(t); FixR
    | sc.SignOf     : Next();
                      IF Result.re.Sign=X.positive THEN XM.Init(Result, X.Ex1, X.Ex0)
                      ELSE XM.Init(Result, X.Ex1, X.Ex0); XM.ExChgSign(Result)
                      END
    | sc.Abs        : Next(); IF Result.re.Sign=X.negative THEN XM.ExChgSign(Result) END;
                      XM.Init(Result, Result.re, X.Ex0)
    | sc.Min        : Comma(); X.ExMin(t, tmp.re, Result.re); FixR
    | sc.Max        : Comma(); X.ExMax(t, tmp.re, Result.re); FixR
    | sc.Conj       : Next(); XM.Conj(Result)
    | sc.Rand       : sc.Get(Token); XMO.random(t); FixR
    ELSE              sc.Mark(X.IllegalOperator); Result := XM.Ex0
  END;
  IF X.ExStatus#X.Okay THEN sc.Mark(X.ExStatus) END
END Factor;


PROCEDURE Powers (VAR Result : XM.ExComplexType);
VAR
  tmp : XM.ExComplexType;
  xtmp : X.ExNumType;

  PROCEDURE Next;
  BEGIN
    sc.Get(Token); Factor(Result);
  END Next;

BEGIN
  Factor(tmp);
  WHILE (Token>=sc.Power) & (Token<=sc.PolarToRect) DO
    CASE Token OF
        sc.Power     : Next(); XM.powerX(tmp, tmp, Result);
      | sc.Root      : Next(); XM.rootX(tmp, Result, tmp);
      | sc.Squared   : sc.Get(Token); XM.ExMult(tmp, tmp, tmp);
      | sc.Cubed     : sc.Get(Token); XM.xtoi(tmp, tmp, 3);
      | sc.Inverse   : sc.Get(Token); XM.ExDiv(tmp, XM.Ex1, tmp);
      | sc.Factorial : sc.Get(Token);
                       XM.factorialX(tmp, X.ExToLongInt(tmp.re));
      | sc.PercentOf : sc.Get(Token);
                       X.ExNumb(0, 1, -1, xtmp);        (* 0.01 *)
                       XM.Init(tmp, xtmp, X.Ex0);
                       XM.ExMult(Result, tmp, Result);
                       Factor(tmp);
                       XM.ExMult(tmp, tmp, Result);
      | sc.PolarToRect:Next(); XM.ToRectangular(Result, tmp.re, Result.re)
      ELSE (* skip *)  sc.Mark(X.IllegalOperator); sc.Get(Token);
    END;
  END;
  Result := tmp;
  IF X.ExStatus#X.Okay THEN sc.Mark(X.ExStatus) END
END Powers;


PROCEDURE Term (VAR Result : XM.ExComplexType);
VAR
  tmp: XM.ExComplexType;
  t : X.ExNumType;

  PROCEDURE Next;
  BEGIN
    sc.Get(Token); Powers(Result);
  END Next;

  PROCEDURE ToCard(Ex : XM.ExComplexType) : INTEGER;
  BEGIN
    RETURN SHORT(X.ExToLongInt(Ex.re));
  END ToCard;

  PROCEDURE Fix;
  BEGIN
    XM.Init(tmp, t, X.Ex0)
  END Fix;

BEGIN
  Powers(tmp);
  WHILE (Token>=sc.Times) & (Token<=sc.nPr) DO
    CASE Token OF
        sc.Times       : Next(); XM.ExMult(tmp, Result, tmp);
      | sc.Divide      : Next(); XM.ExDiv(tmp, tmp, Result);
      | sc.Div         : Next(); XI.ExIntDiv(t, tmp.re, Result.re); Fix
      | sc.Mod         : Next(); XI.ExMod(t, tmp.re, Result.re); Fix
      | sc.And         : Next(); XI.ExAnd(t, tmp.re, Result.re); Fix
      | sc.ShiftRight  : Next(); XI.ExShr(t, tmp.re, ToCard(Result)); Fix
      | sc.AShiftRight : Next(); XI.ExAshr(t, tmp.re, ToCard(Result)); Fix
      | sc.RotateRight : Next(); XI.ExRor(t, tmp.re, ToCard(Result)); Fix
      | sc.ShiftLeft   : Next(); XI.ExShl(t, tmp.re, ToCard(Result)); Fix
      | sc.RotateLeft  : Next(); XI.ExRol(t, tmp.re, ToCard(Result)); Fix
      | sc.ClearBit    : Next(); XI.ExClearBit(t, tmp.re, ToCard(Result)); Fix
      | sc.SetBit      : Next(); XI.ExSetBit(t, tmp.re, ToCard(Result)); Fix
      | sc.ToggleBit   : Next(); XI.ExToggleBit(t, tmp.re, ToCard(Result)); Fix
      | sc.nCr         : Next(); XMO.combinations(t, tmp.re, Result.re); Fix
      | sc.nPr         : Next(); XMO.permutations(t, tmp.re, Result.re); Fix
      ELSE               sc.Mark(X.IllegalOperator); sc.Get(Token)
    END;
  END;
  Result := tmp;
  IF X.ExStatus#X.Okay THEN sc.Mark(X.ExStatus) END  
END Term;


PROCEDURE Expression (VAR Result : XM.ExComplexType);
VAR
  tmp : XM.ExComplexType;
  t    : X.ExNumType;

  PROCEDURE Next(VAR Result : XM.ExComplexType);
  BEGIN
    sc.Get(Token); Term(Result);
  END Next;

  PROCEDURE Fix;
  BEGIN
    XM.Init(tmp, t, X.Ex0)
  END Fix;

BEGIN
  CASE Token OF
      sc.Plus  : Next(tmp);
    | sc.Minus : Next(tmp); XM.ExChgSign(tmp);
      ELSE       Term(tmp)
  END;
  WHILE (Token >= sc.Plus) & (Token <= sc.Store) DO
    CASE Token OF
        sc.Plus  : Next(Result); XM.ExAdd(tmp, tmp, Result);
      | sc.Minus : Next(Result); XM.ExSub(tmp, tmp, Result);
      | sc.Or    : Next(Result); XI.ExOr(t, Result.re, tmp.re); Fix
      | sc.Xor   : Next(Result); XI.ExXor(t, Result.re, tmp.re); Fix
      | sc.Store : sc.Get(Token); 
                   IF Token=sc.Var THEN StoreVariable(sc.var, tmp) 
                   ELSE sc.Mark(cli.IllegalVariable)
                   END;
                   sc.Get(Token)
      ELSE         Term(tmp);
    END;
  END;
  Result := tmp;
  IF X.ExStatus#X.Okay THEN sc.Mark(X.ExStatus) END  
END Expression;


PROCEDURE Evaluate* (VAR arg: ARRAY OF CHAR; VAR Result: XM.ExComplexType): BOOLEAN;
BEGIN
  COPY(arg, CommandLine);      (* remember this string for later        *)
  X.ExStatus := X.Okay;        (* clear out any previous errors         *)
  sc.Init(arg); sc.Get(Token); (* start things off with the first token *)
  Expression(Result);
  IF Token#sc.Empty THEN sc.Mark(cli.IllegalExpression)
  ELSIF X.ExStatus=X.Okay THEN
    set.State.LastAnswer := Result;
    IF ~el.Found(Equations, arg) THEN el.Append(Equations, arg) END
  END;
  RETURN X.ExStatus=X.Okay
END Evaluate;


PROCEDURE Save *;
VAR
  RFile : f.File;
  w: tr.Writer;
  equation, FName : sc.String;
  eqid, res: INTEGER;
BEGIN
  (* save the list equations *)
  COPY(set.UserHome, FName);
  s.Append(EquationFile, FName);
  RFile:=f.New(FName, {f.write}, res);
  IF RFile#NIL THEN
    w:=tr.ConnectWriter(RFile);
    IF w#NIL THEN
      eqid := 1;
      LOOP
        el.ItemIs(Equations, eqid, equation); INC(eqid);
        IF equation[0]=0X THEN EXIT END;
        w.WriteString(equation); w.WriteLn
      END;
      RFile.Register
    END
  END;
  el.Dispose(Equations);  (* clean up the equations *)
  
  COPY(set.UserHome, FName);
  s.Append(ResultFile, FName);
  RFile:=f.New(FName, {f.write}, res);  
  IF  RFile#NIL THEN
    w:=tr.ConnectWriter(RFile);
    IF w#NIL THEN
      eqid := 1;
      LOOP
        el.ItemIs(Results, eqid, equation); INC(eqid);
        IF equation[0]=0X THEN EXIT END;
        w.WriteString(equation); w.WriteLn
      END;
      RFile.Register 
    END
  END;
  el.Dispose(Results);  (* clean up the results *)
END Save;


PROCEDURE Read *;
VAR
  RFile : f.File;
  r: tr.Reader;
  equation, FName : sc.String;
  res: INTEGER;
BEGIN
  el.Init(Equations);     (* empty equation list *)
  el.Init(Results);       (* empty result list *)

  (* get any saved equations *)
  COPY(set.UserHome, FName);
  s.Append(EquationFile, FName);
  RFile:=f.Old(FName, {f.read}, res);
  IF RFile#NIL THEN
    r:=tr.ConnectReader(RFile);
    WHILE (r#NIL) & (r.Res()=f.done) DO
      r.ReadLine(equation);
      el.Append(Equations, equation)
    END;
    RFile.Close
  END;

  (* get any saved results *)
  COPY(set.UserHome, FName);
  s.Append(ResultFile, FName); 
  RFile:=f.Old(FName, {f.read}, res);
  IF RFile#NIL THEN
    r:=tr.ConnectReader(RFile);
    WHILE (r#NIL) & (r.Res()=f.done) DO
      r.ReadLine(equation);
      el.Append(Results, equation)
    END;
    RFile.Close
  END
END Read;


BEGIN
  Token := sc.Empty;
  IF r.Registered THEN EquationFile:=".Equations.bin"; ResultFile:=".Results.bin"
  ELSE EquationFile:=".Equations.bin"; ResultFile:=".Results.bin"
  END;
  Read; Statistics.Clear;
  Rts.RegisterProc(Save)
END Equations.
