MODULE Settings;

 (*
    Settings - Configuration settings.       
    Copyright (C) 1996 Michael Griebling
 
    This module is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
 
    This module is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*)

IMPORT f:=Files, XI:=ExIntegers, XC:=ExComplexLib0, XMO:=ExMathLib0,
       p:=UnixPlot, V:=Variables, X:=ExNumbers, S:=Strings, 
       br:=BinaryRider, Rts, Termination;

CONST
(* $IFNOT Registered *)
  MaxMemory        = 20;   (* 16 memories & 4 other *)
  StateFile        = ".CalculatorState.bin";
(*
  MaxMemory        = 100;
  StateFile        = "S:CalculatorState.bin";
 *)

  MainWin*=0; ConvertWin*=1; FinanceWin*=2; StatisticWin*=3; ProgrammerWin*=4;
  IconWin*=5; CustomWin*=6;
  Equations*=0; Variables*=1; Results*=2;
  Normal*=0; Scientific*=1; Engineering*=2;

TYPE
  PosRec*=RECORD x*,y*: INTEGER END;
  PosType*=ARRAY 8 OF PosRec;
  StateType * = RECORD
                  LocalBase*    : XI.BaseType;
                  DefOpen*      : SET;
                  Positions*    : PosType;
                  DecPoint*     : INTEGER;
                  MaxMemory*    : INTEGER;
                  ActiveList*   : INTEGER;
                  DigSep*       : CHAR;
                  FracSep*      : CHAR;
                  Notation*     : SHORTINT;
                  NumbDigits*   : INTEGER;
                  LastAnswer*   : XC.ExComplexType;
                  DegRadFlag*   : XMO.DegRadType;
                  xDivs*,yDivs* : INTEGER;
                  mode*         : INTEGER;
                  xDec*, yDec*  : INTEGER;
                  xMin*, xMax*  : LONGREAL;
                  yMin*, yMax*  : LONGREAL
                END;


VAR
  State * : StateType;
  StateFName, UserHome - : ARRAY 256 OF CHAR;


PROCEDURE ComplexToStr (A : XC.ExComplexType; VAR Str : ARRAY OF CHAR);
VAR
  Result : ARRAY X.HighBoundsManArray+10 OF CHAR;
  iZero, rZero, found : BOOLEAN;
  ExpWidth: INTEGER; EngFormat: BOOLEAN;
  Aim: X.ExNumType;
  i, ds: LONGINT;
  dp: INTEGER;
  cs: ARRAY 2 OF CHAR;
BEGIN
  IF State.LocalBase=10 THEN
    ds:=3;
    iZero := X.IsZero(A.im); rZero := X.IsZero(A.re);
    EngFormat := State.Notation = Engineering;
    IF State.Notation>Normal THEN ExpWidth:=1 ELSE ExpWidth:=0 END;
    IF iZero OR ~rZero THEN
      X.ExNumToStr(A.re, State.DecPoint, ExpWidth, Str, EngFormat) (* real part *)
    ELSE COPY("", Str)
    END;
    IF ~iZero THEN
      Aim:=A.im;
      IF Aim.Sign = X.positive THEN
        IF ~rZero THEN S.Append("+", Str) END
      ELSE S.Append("-", Str)
      END;
      X.ExAbs(Aim);
      IF X.ExCompare(Aim, X.Ex1) # X.ExEqual THEN
        X.ExNumToStr(Aim, State.DecPoint, ExpWidth, Result, EngFormat);  (* imaginary part *)
        S.Append(Result, Str)
      END;
      S.Append("i", Str)
    END
  ELSE
    IF State.LocalBase=8 THEN ds:=3 ELSE ds:=4 END;
    XI.ExIntToStr(A.re, State.LocalBase, Str)
  END;

  (* add digit separators *)
  cs[1]:=0X;
  S.FindNext(".", Str, 0, found, dp);     (* find location of decimal point *)
  IF ~found THEN dp:=S.Length(Str)
  ELSIF State.DigSep="." THEN Str[dp]:=","
  END;
  IF State.DigSep # 0X THEN
    i:=dp-ds; cs[0]:=State.DigSep;        (* insert whole separator *)
    WHILE i>0 DO
      S.Insert(cs, SHORT(i), Str); DEC(i, ds); INC(dp)
    END
  END;
  IF State.FracSep # 0X THEN
    i:=dp+ds+1; cs[0]:=State.DigSep;      (* insert fraction separator *)
    WHILE i<S.Length(Str) DO
      S.Insert(cs, SHORT(i), Str); INC(i, ds+1)
    END
  END
END ComplexToStr;


PROCEDURE GetPlotState;
BEGIN
  (* get default plot states *)
  State.mode  := p.mode;
  State.xDec  := p.xDec; State.yDec  := p.yDec;
  State.xMin  := p.xMin; State.xMax  := p.xMax;
  State.yMin  := p.yMin; State.yMax  := p.yMax;
  State.xDivs := p.xDivs; State.yDivs := p.yDivs
END GetPlotState;


PROCEDURE Save*;
VAR
  RFile: f.File; res: INTEGER; w: br.Writer;
BEGIN
  (* save calculator state *)
  GetPlotState;
  State.DegRadFlag:=XMO.AngleUnit;
  RFile:=f.New(StateFName, {f.write}, res);
  IF RFile#NIL THEN
    w:=br.ConnectWriter(RFile);
    IF w#NIL THEN
      w.WriteBytes(State, 0, SIZE(StateType));      (* store state *)
      IF V.Write(w) THEN       (* store variables *)
        RFile.Register         (* save file *)
      END
    END
  END
END Save;


PROCEDURE DefaultVariables;
VAR mem: ARRAY 5 OF CHAR; cnt: INTEGER; ok: BOOLEAN;

  PROCEDURE INC (VAR ch: CHAR);
  BEGIN
    ch:=CHR(ORD(ch)+1)
  END INC;
  
BEGIN
  V.Init(State.MaxMemory);
  mem := "M0";
  FOR cnt := 0 TO 15 DO
    IF cnt = 10 THEN mem := "M10" END;
    V.Set(mem, XC.Ex0, ok);
    IF cnt < 10 THEN INC(mem[1]) ELSE INC(mem[2]) END
  END
END DefaultVariables;


PROCEDURE GetWindowPos(VAR Pos: PosType);
BEGIN
  Pos[MainWin].x:=38;        Pos[MainWin].y:=53;
  Pos[ConvertWin].x:=109;    Pos[ConvertWin].y:=248;
  Pos[FinanceWin].x:=138;    Pos[FinanceWin].y:=258;
  Pos[StatisticWin].x:=138;  Pos[StatisticWin].y:=258;
  Pos[ProgrammerWin].x:=138; Pos[ProgrammerWin].y:=258;
  Pos[IconWin].x:=7;         Pos[IconWin].y:=34;
  Pos[CustomWin].x:=137;     Pos[CustomWin].y:=70;
END GetWindowPos;

PROCEDURE Default;
BEGIN
  (* default calculator state *)
  State.LocalBase   := 10;
  State.DecPoint    :=  0;
  State.Notation    := Normal;
  State.NumbDigits  := 16;
  State.DefOpen     := {};
  State.DegRadFlag  := XMO.AngleUnit;
  State.MaxMemory   := MaxMemory;
  State.DigSep      := 0X;             (* no separator *)
  State.FracSep     := 0X;             (* no separator *)
  State.ActiveList  := Equations;

  (* default window positions *)
  GetWindowPos(State.Positions);

  (* get new state -- if available *)
  GetPlotState;

  (* define default Memory locations *)
  DefaultVariables;

  (* set default number of digits *)
  X.SetMaxDigits(State.NumbDigits)
END Default;

PROCEDURE Read*;
VAR
  RFile : f.File; r: br.Reader; res: INTEGER;
BEGIN
  RFile:=f.Old(StateFName, {f.read}, res);
  IF RFile#NIL THEN
    r:=br.ConnectReader(RFile);
    IF r#NIL THEN
      r.ReadBytes(State, 0, SIZE(StateType));
      IF V.Read(r, State.MaxMemory) & (r.Res()=br.done) THEN
        X.SetMaxDigits(State.NumbDigits); XMO.AngleUnit:=State.DegRadFlag
      END
    ELSE Default
    END;
    RFile.Close
  ELSE Default (* no new state available -- define defaults *)
  END;
  p.SetState(State.mode, State.xDec, State.yDec,
             State.xMin, State.xMax,
             State.yMin, State.yMax,
             State.xDivs,State.yDivs)
END Read;


BEGIN
  Rts.GetUserHome(UserHome, "");   (* get the user's home directory *)
  S.Append("/", UserHome);
  StateFName:=UserHome;            (* create the working filename *)
  S.Append(StateFile, StateFName);
  Read;                            (* read external settings *)
  XC.ComplexToStr:=ComplexToStr;   (* install string conversion routine *)
  Termination.RegisterProc(Save);  (* save external settings on shutdown *)
END Settings.





