MODULE UnixExecLists;

 (*
    UnixExecLists - Simple linked list implementation.       
    Copyright (C) 1996 Michael Griebling
 
    This module is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
 
    This module is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*)

TYPE
  String * = POINTER TO ARRAY OF CHAR;
  Node * = POINTER TO NodeDesc;
  List* = RECORD
    head: Node;
    tail: Node
  END;
  NodeDesc * = RECORD
    name: String;
    prev, next: Node;
  END;


PROCEDURE Init * (VAR l : List);
BEGIN
  l.head:=NIL; l.tail:=NIL
END Init;


PROCEDURE Append * (VAR l : List; str : ARRAY OF CHAR);
VAR n, p : Node; sptr : String;
BEGIN
  IF str[0]=0X THEN RETURN END;  (* ignore blanks *)
  NEW(n); NEW(sptr, LEN(str));
  COPY(str, sptr^); n.name := sptr; n.next:=NIL;
  IF l.tail=NIL THEN l.head:=n; l.tail:=n; n.prev:=NIL
  ELSE p:=l.tail; p.next:=n; (* update previous node *)
    n.prev:=p;               (* link in the new node *)
    l.tail:=n                (* new tail defined *)
  END
END Append;


PROCEDURE Found * (l : List; str : ARRAY OF CHAR) : BOOLEAN;
VAR n: Node;
BEGIN
  n:=l.head;
  WHILE (n#NIL) & (n.name^#str) DO n:=n.next END;
  RETURN n#NIL
END Found;


PROCEDURE ItemIs * (l : List; item : INTEGER; VAR str : ARRAY OF CHAR);
(* Get the itemth list element where the first list item is numbered 1 *)
VAR n : Node;
BEGIN IF l.head=NIL THEN COPY("", str); RETURN END;
  n := l.head;
  WHILE (n # NIL) & (item > 1) DO n := n.next; DEC(item) END;
  IF n = NIL THEN COPY("", str)
  ELSE COPY(n.name^, str)
  END
END ItemIs;


PROCEDURE Count * (l : List) : LONGINT;
(* Return the number of items in the given list *)
VAR n : Node; total: LONGINT;
BEGIN IF l.head=NIL THEN RETURN 0 END;
  n := l.head; total:=1;
  WHILE n # NIL DO n := n.next; INC(total) END;
  RETURN total
END Count;


PROCEDURE Dispose * (VAR l : List);
BEGIN
  Init(l)    (* clear out the list *)
END Dispose;


END UnixExecLists.





