MODULE Variables;

 (*
    Variables - Variable storage/recall.       
    Copyright (C) 1996 Michael Griebling
 
    This module is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
 
    This module is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*)

IMPORT C:= ExComplexLib0, el:= UnixExecLists, str:= Strings,
       F:= BinaryRider, sys:= SYSTEM;

CONST
  VarSize* = 16;

TYPE
  VarType = RECORD
    Name: ARRAY VarSize OF CHAR; (* names have maximum of 15 characters *)
    Value: C.ExComplexType
  END;
  VarPtr = POINTER TO ARRAY OF VarType;
  VarRecord = RECORD
    MaxVars, Defined: LONGINT;
    Vars: VarPtr
  END;
  IteratorType*=PROCEDURE(name, arg: ARRAY OF CHAR);

CONST
  NotFound = -1;

VAR
  Var: VarRecord;      (* all defined variables *)
  Variables-: el.List; (* list passed to scroll window *)

  (* $CopyArrays- *)
  PROCEDURE Find (var : ARRAY OF CHAR) : LONGINT;
  VAR pos: LONGINT;
  BEGIN pos := 0;
    WHILE pos <= Var.Defined DO
      IF var = Var.Vars[pos].Name THEN RETURN pos
      ELSE INC(pos)
      END
    END;
    RETURN NotFound
  END Find;

  PROCEDURE Max * () : LONGINT;
  BEGIN
    RETURN Var.MaxVars
  END Max;

  PROCEDURE Defined * () : LONGINT;
  BEGIN
    RETURN Var.Defined+1
  END Defined;

  (* $CopyArrays- *)
  PROCEDURE Set * (var : ARRAY OF CHAR; value : C.ExComplexType; VAR ok: BOOLEAN);
    VAR pos: LONGINT;
  BEGIN
    pos := Find(var); ok := TRUE;
    IF pos = NotFound THEN (* this is a new variable *)
      IF Var.Defined+1 < Var.MaxVars THEN INC(Var.Defined)
      ELSE ok := FALSE; RETURN (* no more variables *)
      END;
      pos := Var.Defined;
      COPY(var, Var.Vars[pos].Name);  (* set the name *)
    END;

    (* set the value of the variable *)
    Var.Vars[pos].Value := value
  END Set;

  (* $CopyArrays- *)
  PROCEDURE Get * (var : ARRAY OF CHAR; VAR value : C.ExComplexType; VAR ok: BOOLEAN);
    VAR pos: LONGINT;
  BEGIN
    pos := Find(var);
    IF pos = NotFound THEN ok := FALSE; value := C.Ex0
    ELSE ok := TRUE; value := Var.Vars[pos].Value
    END
  END Get;
  
  PROCEDURE Delete * (var : ARRAY OF CHAR);
    VAR i, pos: LONGINT;
  BEGIN
    pos := Find(var);
    IF pos # NotFound THEN
      DEC(Var.Defined);
      FOR i:=pos TO Var.Defined DO
        Var.Vars[i]:=Var.Vars[i+1]
      END
    END
  END Delete;

  PROCEDURE GetList * () : LONGINT;
    VAR cnt: LONGINT; s, n: ARRAY 256 OF CHAR;
  BEGIN
    el.Dispose(Variables);  (* clear out old list *)
    FOR cnt := 0 TO Var.Defined DO
      COPY(Var.Vars[cnt].Name, s);   (* copy over the name *)
      str.Append(" : ", s);
      C.ComplexToStr(Var.Vars[cnt].Value, n);
      str.Append(n, s);
      el.Append(Variables, s)
    END;
    RETURN sys.ADR(Variables)
  END GetList;

  PROCEDURE Iterate * (IterFunc: IteratorType);
    VAR cnt: LONGINT; n: ARRAY 256 OF CHAR;
  BEGIN
    FOR cnt := 0 TO Var.Defined DO
      C.ComplexToStr(Var.Vars[cnt].Value, n);
      IterFunc(Var.Vars[cnt].Name, n)
    END
  END Iterate;
  
  PROCEDURE Write * (VAR f : F.Writer) : BOOLEAN;
  BEGIN
    f.WriteNum(Var.MaxVars);   (* write variable header *)
    f.WriteNum(Var.Defined); 
    IF Var.MaxVars > 0 THEN     (* write the data *)   
      f.WriteBytes(Var.Vars^, 0, Var.MaxVars*SIZE(VarType))
    END;
    RETURN f.Res()=F.done
  END Write;

  PROCEDURE Read * (VAR f : F.Reader; NewMax : INTEGER) : BOOLEAN;
  BEGIN
    f.ReadNum(Var.MaxVars);     (* read variable header *)
    f.ReadNum(Var.Defined);
    IF NewMax>0 THEN Var.MaxVars:=NewMax END;
    IF Var.MaxVars > 0 THEN
      NEW(Var.Vars, Var.MaxVars);
      f.ReadBytes(Var.Vars^, 0, Var.MaxVars*SIZE(VarType))  (* read the data *)
    END;
    RETURN f.Res()=F.done
  END Read;

  PROCEDURE Init * (Max: INTEGER);
  BEGIN
    IF Max>0 THEN NEW(Var.Vars, Max) ELSE Var.Vars:= NIL END;
    Var.MaxVars := Max; Var.Defined := -1;
    el.Init(Variables)
  END Init;

  PROCEDURE Dispose *;
  BEGIN
    el.Dispose(Variables); Init(0)
  END Dispose;

BEGIN
  Init(0)
END Variables.


